@@ -59,6 +59,102 @@ static uint8_t esm_cause_from_gtp(uint8_t gtp_cause)
     return OGS_NAS_ESM_CAUSE_NETWORK_FAILURE;
 }
 
+static void gtp_remote_peer_timeout(ogs_gtp_xact_t *xact, void *data)
+{
+    char buf[OGS_ADDRSTRLEN];
+    mme_bearer_t *bearer = NULL;
+    uint8_t type;
+
+    ogs_assert(xact);
+    type = xact->seq[xact->step-1].type;
+
+    ogs_error("[%d] %s Peer Timeout for step %d type %d peer [%s]:%d",
+            xact->xid,
+            xact->org == OGS_GTP_LOCAL_ORIGINATOR ? "LOCAL " : "REMOTE",
+            xact->step, type,
+            OGS_ADDR(&xact->gnode->addr, buf),
+            OGS_PORT(&xact->gnode->addr));
+
+    ogs_assert(data);
+    bearer = mme_bearer_find_by_id(OGS_POINTER_TO_UINT(data));
+    if (!bearer) {
+        ogs_error("Bearer has already been removed [%d]", type);
+        return;
+    }
+
+    /*
+     * Issues #3240
+     *
+     * SMF->SGW-C->MME: First Update Bearer Request
+     * MME->UE:         First Modify EPS bearer context request
+     * SMF->SGW-C->MME: Second Update Bearer Request
+     * MME->UE:         Second Modify EPS bearer context request
+     * UE->MME:         First Modify EPS bearer context accept
+     * MME->SGW-C->SMF: First Update Bearer Response
+     * UE->MME:         Second Modify EPS bearer context accept
+     * MME->SGW-C->SMF: Second Update Bearer Response
+     */
+    switch (type) {
+    case OGS_GTP2_UPDATE_BEARER_REQUEST_TYPE:
+        /*
+         * In this case, a timeout occurs while waiting
+         * for Modify EPS bearer context accept from UE.
+         *
+         * If the UE does not send a Modify EPS bearer context accept,
+         * the MME fails to send an Update Bearer Response.
+         *
+         * Therefore, we need to delete the Transaction Node
+         * that was managed by the Bearer Context from the List.
+         */
+        if (ogs_list_exists(
+                    &bearer->update.xact_list,
+                    &xact->to_update_node) == true) {
+            ogs_error("Bearer-ID [%d] removed from the list", bearer->id);
+            ogs_list_remove(&bearer->update.xact_list, &xact->to_update_node);
+        } else {
+            ogs_error("[%d] %s HAVE ALREADY BEEN REMOVED "
+                    "for step %d type %d peer [%s]:%d",
+                    xact->xid,
+                    xact->org == OGS_GTP_LOCAL_ORIGINATOR ? "LOCAL " : "REMOTE",
+                    xact->step, type,
+                    OGS_ADDR(&xact->gnode->addr, buf),
+                    OGS_PORT(&xact->gnode->addr));
+        }
+        break;
+#if 0
+    case OGS_GTP2_UPDATE_BEARER_RESPONSE_TYPE:
+        /*
+         * The following is the case where the UE sends
+         * Modify EPS bearer context accept to the MME.
+         *
+         * In this case, the MME sends Update Bearer Response
+         * to SGW-C and deletes the Transaction Node.
+         *
+         * Therefore, there is no need to delete the Transaction Node
+         * from the list managed by the Bearer Context here.
+         */
+        if (ogs_list_exists(
+                    &bearer->update.xact_list,
+                    &xact->to_update_node) == true) {
+            ogs_error("[%d] %s SHOULD HAVE REMOVED "
+                    "for step %d type %d peer [%s]:%d",
+                    xact->xid,
+                    xact->org == OGS_GTP_LOCAL_ORIGINATOR ? "LOCAL " : "REMOTE",
+                    xact->step, type,
+                    OGS_ADDR(&xact->gnode->addr, buf),
+                    OGS_PORT(&xact->gnode->addr));
+        } else {
+            ogs_error("Not existed");
+        }
+        break;
+#endif
+    default:
+        ogs_fatal("Unknown type[%d]", type);
+        ogs_assert_if_reached();
+        break;
+    }
+}
+
 void mme_s11_handle_echo_request(
         ogs_gtp_xact_t *xact, ogs_gtp2_echo_request_t *req)
 {
@@ -94,6 +190,7 @@ void mme_s11_handle_create_session_response(
     mme_sess_t *sess = NULL;
     mme_ue_t *mme_ue = NULL;
     sgw_ue_t *source_ue = NULL, *target_ue = NULL;
+    enb_ue_t *enb_ue = NULL;
     ogs_session_t *session = NULL;
     ogs_gtp2_bearer_qos_t bearer_qos;
     ogs_gtp2_ambr_t *ambr = NULL;
@@ -103,16 +200,16 @@ void mme_s11_handle_create_session_response(
     ogs_assert(rsp);
 
     ogs_debug("Create Session Response");
-    MME_UE_LIST_CHECK;
 
     /********************
      * Check Transaction
      ********************/
     ogs_assert(xact);
     create_action = xact->create_action;
-    sess = xact->data;
+    sess = mme_sess_find_by_id(OGS_POINTER_TO_UINT(xact->data));
     ogs_assert(sess);
-    mme_ue = mme_ue_cycle(sess->mme_ue);
+
+    mme_ue = mme_ue_find_by_id(sess->mme_ue_id);
 
     rv = ogs_gtp_xact_commit(xact);
     if (rv != OGS_OK) {
@@ -124,11 +221,12 @@ void mme_s11_handle_create_session_response(
         ogs_error("MME-UE Context has already been removed");
         return;
     }
-    source_ue = sgw_ue_cycle(mme_ue->sgw_ue);
+    source_ue = sgw_ue_find_by_id(mme_ue->sgw_ue_id);
     ogs_assert(source_ue);
+    enb_ue = enb_ue_find_by_id(mme_ue->enb_ue_id);
 
     if (create_action == OGS_GTP_CREATE_IN_PATH_SWITCH_REQUEST) {
-        target_ue = sgw_ue_cycle(source_ue->target_ue);
+        target_ue = sgw_ue_find_by_id(source_ue->target_ue_id);
         ogs_assert(target_ue);
     } else {
         target_ue = source_ue;
@@ -159,7 +257,7 @@ void mme_s11_handle_create_session_response(
         if (create_action == OGS_GTP_CREATE_IN_ATTACH_REQUEST) {
             ogs_error("[%s] Attach reject [Cause:%d]",
                     mme_ue->imsi_bcd, session_cause);
-            r = nas_eps_send_attach_reject(mme_ue,
+            r = nas_eps_send_attach_reject(enb_ue, mme_ue,
                     OGS_NAS_EMM_CAUSE_NETWORK_FAILURE,
                     OGS_NAS_ESM_CAUSE_NETWORK_FAILURE);
             ogs_expect(r == OGS_OK);
@@ -180,38 +278,21 @@ void mme_s11_handle_create_session_response(
         cause_value = OGS_GTP2_CAUSE_CONDITIONAL_IE_MISSING;
     }
 
-    if (create_action == OGS_GTP_CREATE_IN_PATH_SWITCH_REQUEST) {
-
-        /* No need S5C TEID in PathSwitchRequest */
-
-    } else {
-
+    switch (create_action) {
+    case OGS_GTP_CREATE_IN_PATH_SWITCH_REQUEST:
+        /* No need for PAA or S5C TEID in PathSwitchRequest */
+        break;
+    case OGS_GTP_CREATE_IN_TRACKING_AREA_UPDATE:
+        /* No need for PAA or S5C TEID in 2G->4G mobility, it was already provided by SGSN peer */
+        break;
+    default:
         if (rsp->pgw_s5_s8__s2a_s2b_f_teid_for_pmip_based_interface_or_for_gtp_based_control_plane_interface.presence == 0) {
             ogs_error("[%s] No S5C TEID [Cause:%d]",
                     mme_ue->imsi_bcd, session_cause);
             cause_value = OGS_GTP2_CAUSE_CONDITIONAL_IE_MISSING;
         }
 
-    }
-
-    if (create_action == OGS_GTP_CREATE_IN_PATH_SWITCH_REQUEST) {
-
-        /* No need S5C TEID in PathSwitchRequest */
-
-    } else {
-
-        if (rsp->pdn_address_allocation.presence) {
-            ogs_paa_t paa;
-
-            memcpy(&paa, rsp->pdn_address_allocation.data,
-                    rsp->pdn_address_allocation.len);
-
-            if (!OGS_PDU_SESSION_TYPE_IS_VALID(paa.session_type)) {
-                ogs_error("[%s] Unknown PDN Type [Session:%u, Cause:%d]",
-                        mme_ue->imsi_bcd, paa.session_type, session_cause);
-                cause_value = OGS_GTP2_CAUSE_MANDATORY_IE_INCORRECT;
-            }
-        } else {
+        if (rsp->pdn_address_allocation.presence == 0) {
             ogs_error("[%s] No PDN Address Allocation [Cause:%d]",
                     mme_ue->imsi_bcd, session_cause);
             cause_value = OGS_GTP2_CAUSE_CONDITIONAL_IE_MISSING;
@@ -227,11 +308,18 @@ void mme_s11_handle_create_session_response(
         if (create_action == OGS_GTP_CREATE_IN_ATTACH_REQUEST) {
             ogs_error("[%s] Attach reject [Cause:%d]",
                     mme_ue->imsi_bcd, session_cause);
-            r = nas_eps_send_attach_reject(mme_ue,
+            r = nas_eps_send_attach_reject(enb_ue, mme_ue,
                     OGS_NAS_EMM_CAUSE_NETWORK_FAILURE,
                     OGS_NAS_ESM_CAUSE_NETWORK_FAILURE);
             ogs_expect(r == OGS_OK);
             ogs_assert(r != OGS_ERROR);
+        } else if (create_action == OGS_GTP_CREATE_IN_TRACKING_AREA_UPDATE) {
+            ogs_error("[%s] TAU reject [Cause:%d]",
+                    mme_ue->imsi_bcd, session_cause);
+            r = nas_eps_send_tau_reject(enb_ue, mme_ue,
+                    OGS_NAS_EMM_CAUSE_NETWORK_FAILURE);
+            ogs_expect(r == OGS_OK);
+            ogs_assert(r != OGS_ERROR);
         }
         mme_send_delete_session_or_mme_ue_context_release(mme_ue);
         return;
@@ -259,7 +347,7 @@ void mme_s11_handle_create_session_response(
                     mme_ue->imsi_bcd, bearer_cause);
             if (create_action == OGS_GTP_CREATE_IN_ATTACH_REQUEST) {
                 ogs_error("[%s] Attach reject", mme_ue->imsi_bcd);
-                r = nas_eps_send_attach_reject(mme_ue,
+                r = nas_eps_send_attach_reject(enb_ue, mme_ue,
                         OGS_NAS_EMM_CAUSE_NETWORK_FAILURE,
                         OGS_NAS_ESM_CAUSE_NETWORK_FAILURE);
                 ogs_expect(r == OGS_OK);
@@ -279,7 +367,7 @@ void mme_s11_handle_create_session_response(
         ogs_error("[%s] GTP Cause [VALUE:%d]", mme_ue->imsi_bcd, session_cause);
         if (create_action == OGS_GTP_CREATE_IN_ATTACH_REQUEST) {
             ogs_error("[%s] Attach reject", mme_ue->imsi_bcd);
-            r = nas_eps_send_attach_reject(mme_ue,
+            r = nas_eps_send_attach_reject(enb_ue, mme_ue,
                     OGS_NAS_EMM_CAUSE_NETWORK_FAILURE,
                     OGS_NAS_ESM_CAUSE_NETWORK_FAILURE);
             ogs_expect(r == OGS_OK);
@@ -342,6 +430,10 @@ void mme_s11_handle_create_session_response(
 
         ogs_debug("    ENB_S1U_TEID[%d] SGW_S1U_TEID[%d] PGW_S5U_TEID[%d]",
             bearer->enb_s1u_teid, bearer->sgw_s1u_teid, bearer->pgw_s5u_teid);
+
+        if (create_action == OGS_GTP_CREATE_IN_TRACKING_AREA_UPDATE &&
+            !OGS_FSM_CHECK(&bearer->sm, esm_state_active))
+            OGS_FSM_TRAN(&bearer->sm, esm_state_active);
     }
 
     /* Bearer Level QoS */
@@ -372,11 +464,24 @@ void mme_s11_handle_create_session_response(
 
     /* PDN Addresss Allocation */
     if (rsp->pdn_address_allocation.presence) {
-        memcpy(&session->paa, rsp->pdn_address_allocation.data,
+        memcpy(&sess->paa, rsp->pdn_address_allocation.data,
                 rsp->pdn_address_allocation.len);
-        session->session_type = session->paa.session_type;
-        ogs_assert(OGS_OK ==
-                ogs_paa_to_ip(&session->paa, &session->ue_ip));
+        /*
+         * Issue #3209
+         *
+         * The Session-Type in the Subscriber DB should not be changed
+         * in case the UE can change the PDN-Type for the APN.
+         * (e.g IPv4v6 -> IPv4 -> IPv4v6)
+         *
+         * For resolving this problem,
+         * session->session_type and session->ue_ip should not be modified.
+         *
+         * Therefore, the code below will be deleted.
+         */
+#if 0 /* WILL BE DELETED */
+        session->session_type = sess->paa.session_type;
+        ogs_assert(OGS_OK == ogs_paa_to_ip(&sess->paa, &session->ue_ip));
+#endif
     }
 
     /* ePCO */
@@ -415,21 +520,22 @@ void mme_s11_handle_create_session_response(
         mme_csmap_t *csmap = mme_csmap_find_by_tai(&mme_ue->tai);
         mme_ue->csmap = csmap;
 
-        if (csmap) {
-            ogs_assert(OGS_PDU_SESSION_TYPE_IS_VALID(
-                        session->paa.session_type));
-            ogs_assert(OGS_OK ==
-                sgsap_send_location_update_request(mme_ue));
-        } else {
-            ogs_assert(OGS_PDU_SESSION_TYPE_IS_VALID(
-                        session->paa.session_type));
+        if (!csmap ||
+            mme_ue->network_access_mode ==
+                OGS_NETWORK_ACCESS_MODE_ONLY_PACKET ||
+            mme_ue->nas_eps.attach.value ==
+                OGS_NAS_ATTACH_TYPE_EPS_ATTACH) {
             r = nas_eps_send_attach_accept(mme_ue);
             ogs_expect(r == OGS_OK);
             ogs_assert(r != OGS_ERROR);
+        } else {
+            ogs_assert(OGS_OK == sgsap_send_location_update_request(mme_ue));
         }
 
+    } else if (create_action == OGS_GTP_CREATE_IN_TRACKING_AREA_UPDATE) {
+        /* 3GPP TS 23.401 D.3.6 step 13, 14: */
+        mme_s6a_send_ulr(enb_ue, mme_ue);
     } else if (create_action == OGS_GTP_CREATE_IN_UPLINK_NAS_TRANSPORT) {
-        ogs_assert(OGS_PDU_SESSION_TYPE_IS_VALID(session->paa.session_type));
         r = nas_eps_send_activate_default_bearer_context_request(
                 bearer, create_action);
         ogs_expect(r == OGS_OK);
@@ -467,14 +573,14 @@ void mme_s11_handle_modify_bearer_response(
     ogs_assert(rsp);
 
     ogs_debug("Modify Bearer Response");
-    MME_UE_LIST_CHECK;
 
     /********************
      * Check Transaction
      ********************/
     ogs_assert(xact);
     modify_action = xact->modify_action;
-    mme_ue = mme_ue_cycle(xact->data);
+
+    mme_ue = mme_ue_find_by_id(OGS_POINTER_TO_UINT(xact->data));
 
     rv = ogs_gtp_xact_commit(xact);
     if (rv != OGS_OK) {
@@ -486,7 +592,7 @@ void mme_s11_handle_modify_bearer_response(
         ogs_error("MME-UE Context has already been removed");
         return;
     }
-    sgw_ue = sgw_ue_cycle(mme_ue->sgw_ue);
+    sgw_ue = sgw_ue_find_by_id(mme_ue->sgw_ue_id);
     ogs_assert(sgw_ue);
 
     /************************
@@ -575,11 +681,11 @@ void mme_s11_handle_delete_session_response(
     sgw_ue_t *source_ue = NULL, *target_ue = NULL;
     mme_sess_t *sess = NULL;
     mme_ue_t *mme_ue = NULL;
+    enb_ue_t *enb_ue = NULL;
 
     ogs_assert(rsp);
 
     ogs_debug("Delete Session Response");
-    MME_UE_LIST_CHECK;
 
     /********************
      * Check Transaction
@@ -587,9 +693,11 @@ void mme_s11_handle_delete_session_response(
     ogs_assert(xact);
     action = xact->delete_action;
     ogs_assert(action);
-    sess = xact->data;
+    sess = mme_sess_find_by_id(OGS_POINTER_TO_UINT(xact->data));
     ogs_assert(sess);
-    mme_ue = mme_ue_cycle(sess->mme_ue);
+
+    ogs_debug("delete_session_response - xact:%p, sess:%p", xact, sess);
+    mme_ue = mme_ue_find_by_id(sess->mme_ue_id);
 
     rv = ogs_gtp_xact_commit(xact);
     if (rv != OGS_OK) {
@@ -601,11 +709,11 @@ void mme_s11_handle_delete_session_response(
         ogs_error("MME-UE Context has already been removed");
         return;
     }
-    target_ue = sgw_ue_cycle(mme_ue->sgw_ue);
+    target_ue = sgw_ue_find_by_id(mme_ue->sgw_ue_id);
     ogs_assert(target_ue);
 
     if (action == OGS_GTP_DELETE_IN_PATH_SWITCH_REQUEST) {
-        source_ue = sgw_ue_cycle(target_ue->source_ue);
+        source_ue = sgw_ue_find_by_id(target_ue->source_ue_id);
         if (!source_ue) /* InterRAT to 2G/3G (SGSN) case: */
              source_ue = target_ue;
         ogs_assert(source_ue);
@@ -654,6 +762,8 @@ void mme_s11_handle_delete_session_response(
     ogs_assert(target_ue);
     ogs_assert(source_ue);
 
+    enb_ue = enb_ue_find_by_id(mme_ue->enb_ue_id);
+
     ogs_debug("    MME_S11_TEID[%d] SGW_S11_TEID[%d]",
             mme_ue->mme_s11_teid, source_ue->sgw_s11_teid);
 
@@ -663,7 +773,7 @@ void mme_s11_handle_delete_session_response(
          * of the detach accept from UE */
     } else if (action == OGS_GTP_DELETE_SEND_AUTHENTICATION_REQUEST) {
         if (mme_sess_count(mme_ue) == 1) /* Last Session */ {
-            mme_s6a_send_air(mme_ue, NULL);
+            mme_s6a_send_air(enb_ue, mme_ue, NULL);
         }
 
     } else if (action == OGS_GTP_DELETE_SEND_DETACH_ACCEPT) {
@@ -691,14 +801,9 @@ void mme_s11_handle_delete_session_response(
     } else if (action == OGS_GTP_DELETE_SEND_RELEASE_WITH_UE_CONTEXT_REMOVE) {
         if (mme_sess_count(mme_ue) == 1) /* Last Session */ {
             if (ECM_IDLE(mme_ue)) {
-                if (mme_ue->location_updated_but_not_canceled_yet == true) {
-                    mme_s6a_send_pur(mme_ue);
-                } else {
-                    mme_ue_remove(mme_ue);
-                }
+                mme_ue_remove(mme_ue);
             } else {
-                ogs_assert(mme_ue->enb_ue);
-                r = s1ap_send_ue_context_release_command(mme_ue->enb_ue,
+                r = s1ap_send_ue_context_release_command(enb_ue,
                     S1AP_Cause_PR_nas, S1AP_CauseNas_normal_release,
                     S1AP_UE_CTX_REL_UE_CONTEXT_REMOVE, 0);
                 ogs_expect(r == OGS_OK);
@@ -709,17 +814,11 @@ void mme_s11_handle_delete_session_response(
     } else if (action ==
             OGS_GTP_DELETE_SEND_RELEASE_WITH_S1_REMOVE_AND_UNLINK) {
         if (mme_sess_count(mme_ue) == 1) /* Last Session */ {
-            enb_ue_t *enb_ue = NULL;
-
-            enb_ue = enb_ue_cycle(mme_ue->enb_ue);
-            if (enb_ue) {
-                r = s1ap_send_ue_context_release_command(enb_ue,
-                        S1AP_Cause_PR_nas, S1AP_CauseNas_normal_release,
-                        S1AP_UE_CTX_REL_S1_REMOVE_AND_UNLINK, 0);
-                ogs_expect(r == OGS_OK);
-                ogs_assert(r != OGS_ERROR);
-            } else
-                ogs_error("ENB-S1 Context has already been removed");
+            r = s1ap_send_ue_context_release_command(enb_ue,
+                    S1AP_Cause_PR_nas, S1AP_CauseNas_normal_release,
+                    S1AP_UE_CTX_REL_S1_REMOVE_AND_UNLINK, 0);
+            ogs_expect(r == OGS_OK);
+            ogs_assert(r != OGS_ERROR);
         }
 
     } else if (action == OGS_GTP_DELETE_HANDLE_PDN_CONNECTIVITY_REQUEST) {
@@ -728,7 +827,7 @@ void mme_s11_handle_delete_session_response(
                     &mme_ue->pdn_connectivity_request);
             if (rv != OGS_OK) {
                 ogs_error("nas_eps_send_emm_to_esm() failed");
-                r = nas_eps_send_attach_reject(mme_ue,
+                r = nas_eps_send_attach_reject(enb_ue, mme_ue,
                         OGS_NAS_EMM_CAUSE_PROTOCOL_ERROR_UNSPECIFIED,
                         OGS_NAS_ESM_CAUSE_PROTOCOL_ERROR_UNSPECIFIED);
                 ogs_expect(r == OGS_OK);
@@ -775,7 +874,7 @@ void mme_s11_handle_create_bearer_request(
     ogs_assert(xact);
     ogs_assert(req);
 
-    ogs_debug("Create Bearer Response");
+    ogs_debug("Create Bearer Request");
 
     /***********************
      * Check MME-UE Context
@@ -786,7 +885,7 @@ void mme_s11_handle_create_bearer_request(
         ogs_error("No Context in TEID");
         cause_value = OGS_GTP2_CAUSE_CONTEXT_NOT_FOUND;
     } else {
-        sgw_ue = sgw_ue_cycle(mme_ue->sgw_ue);
+        sgw_ue = sgw_ue_find_by_id(mme_ue->sgw_ue_id);
         ogs_assert(sgw_ue);
 
         if (req->linked_eps_bearer_id.presence == 0) {
@@ -859,12 +958,25 @@ void mme_s11_handle_create_bearer_request(
     ogs_debug("    MME_S11_TEID[%d] SGW_S11_TEID[%d]",
             mme_ue->mme_s11_teid, sgw_ue->sgw_s11_teid);
 
+    /*
+     * DEPRECATED : Issues #3072
+     *
+     * PTI 0 is set here to prevent a InitialContextSetupRequest message
+     * with a PTI of 0 from being created when the Create Bearer Request occurs
+     * and InitialContextSetupRequest occurs.
+     *
+     * If you implement the creation of a dedicated bearer
+     * in the ESM procedure reqeusted by the UE,
+     * you will need to refactor the part that sets the PTI.
+     */
+#if 0
     /* Set PTI */
     sess->pti = OGS_NAS_PROCEDURE_TRANSACTION_IDENTITY_UNASSIGNED;
     if (req->procedure_transaction_id.presence) {
         sess->pti = req->procedure_transaction_id.u8;
         ogs_debug("    PTI[%d]", sess->pti);
     }
+#endif
 
     /* Data Plane(UL) : SGW-S1U */
     sgw_s1u_teid = req->bearer_contexts.s1_u_enodeb_f_teid.data;
@@ -918,7 +1030,8 @@ void mme_s11_handle_create_bearer_request(
      * If GTP-xact Holding timer is expired,
      * OLD bearer->xact memory will be automatically removed.
      */
-    bearer->create.xact = xact;
+    ogs_assert(xact->id >= OGS_MIN_POOL_ID && xact->id <= OGS_MAX_POOL_ID);
+    bearer->create.xact_id = xact->id;
 
     /* Before Activate DEDICATED bearer, check DEFAULT bearer status */
     default_bearer = mme_default_bearer_in_sess(sess);
@@ -930,7 +1043,7 @@ void mme_s11_handle_create_bearer_request(
     if (OGS_FSM_CHECK(&default_bearer->sm, esm_state_active)) {
         if (ECM_IDLE(mme_ue)) {
             MME_STORE_PAGING_INFO(mme_ue,
-                MME_PAGING_TYPE_CREATE_BEARER, bearer);
+                MME_PAGING_TYPE_CREATE_BEARER, bearer->id);
             r = s1ap_send_paging(mme_ue, S1AP_CNDomain_ps);
             ogs_expect(r == OGS_OK);
             ogs_assert(r != OGS_ERROR);
@@ -974,7 +1087,7 @@ void mme_s11_handle_update_bearer_request(
         ogs_error("No Context in TEID");
         cause_value = OGS_GTP2_CAUSE_CONTEXT_NOT_FOUND;
     } else {
-        sgw_ue = sgw_ue_cycle(mme_ue->sgw_ue);
+        sgw_ue = sgw_ue_find_by_id(mme_ue->sgw_ue_id);
         ogs_assert(sgw_ue);
 
         if (req->bearer_contexts.presence == 0) {
@@ -1009,7 +1122,7 @@ void mme_s11_handle_update_bearer_request(
     ogs_assert(sgw_ue);
 
     ogs_assert(bearer);
-    sess = bearer->sess;
+    sess = mme_sess_find_by_id(bearer->sess_id);
     ogs_assert(sess);
 
     ogs_debug("    MME_S11_TEID[%d] SGW_S11_TEID[%d]",
@@ -1023,13 +1136,31 @@ void mme_s11_handle_update_bearer_request(
     }
 
     /*
-     * Save Transaction. It will be handled after EMM-attached
+     * Issues #3240
      *
-     * You should not remove OLD bearer->xact.
-     * If GTP-xact Holding timer is expired,
-     * OLD bearer->xact memory will be automatically removed.
+     * SMF->SGW-C->MME: First Update Bearer Request
+     * MME->UE:         First Modify EPS bearer context request
+     * SMF->SGW-C->MME: Second Update Bearer Request
+     * MME->UE:         Second Modify EPS bearer context request
+     * UE->MME:         First Modify EPS bearer context accept
+     * MME->SGW-C->SMF: First Update Bearer Response
+     * UE->MME:         Second Modify EPS bearer context accept
+     * MME->SGW-C->SMF: Second Update Bearer Response
+     *
+     * If the UE does not send a Modify EPS bearer context accept,
+     * the MME cannot send an Update Bearer Response to the SGW-C.
+     *
+     * In this case, peer timeout occurs, and a callback function
+     * is registered as follows to free memory.
+     *
+     * Also, as shown above, multiple Update Bearer Request/Response can occur,
+     * so we manage the Transaction Node as a list within the Bearer Context.
      */
-    bearer->update.xact = xact;
+
+    xact->peer_cb = gtp_remote_peer_timeout;
+    xact->peer_data = OGS_UINT_TO_POINTER(bearer->id);
+
+    ogs_list_add(&bearer->update.xact_list, &xact->to_update_node);
 
     if (req->bearer_contexts.bearer_level_qos.presence == 1) {
         /* Bearer QoS */
@@ -1064,11 +1195,19 @@ void mme_s11_handle_update_bearer_request(
         req->bearer_contexts.tft.presence == 1) {
         if (ECM_IDLE(mme_ue)) {
             MME_STORE_PAGING_INFO(mme_ue,
-                MME_PAGING_TYPE_UPDATE_BEARER, bearer);
+                MME_PAGING_TYPE_UPDATE_BEARER, bearer->id);
             r = s1ap_send_paging(mme_ue, S1AP_CNDomain_ps);
             ogs_expect(r == OGS_OK);
             ogs_assert(r != OGS_ERROR);
         } else {
+            /*
+             * MME must wait for Modify Bearer Context Accept
+             * before sending Update Bearer Response,
+             * To check this, start a peer timer to check it.
+             */
+            ogs_timer_start(xact->tm_peer,
+                    ogs_local_conf()->time.message.gtp.t3_response_duration);
+
             MME_CLEAR_PAGING_INFO(mme_ue);
             r = nas_eps_send_modify_bearer_context_request(bearer,
                     req->bearer_contexts.bearer_level_qos.presence,
@@ -1121,7 +1260,7 @@ void mme_s11_handle_delete_bearer_request(
         ogs_error("No Context in TEID");
         cause_value = OGS_GTP2_CAUSE_CONTEXT_NOT_FOUND;
     } else {
-        sgw_ue = sgw_ue_cycle(mme_ue->sgw_ue);
+        sgw_ue = sgw_ue_find_by_id(mme_ue->sgw_ue_id);
         ogs_assert(sgw_ue);
 
         if (req->linked_eps_bearer_id.presence == 1) {
@@ -1181,7 +1320,7 @@ void mme_s11_handle_delete_bearer_request(
     ogs_assert(sgw_ue);
 
     ogs_assert(bearer);
-    sess = bearer->sess;
+    sess = mme_sess_find_by_id(bearer->sess_id);
     ogs_assert(sess);
 
     ogs_debug("    MME_S11_TEID[%d] SGW_S11_TEID[%d]",
@@ -1201,11 +1340,12 @@ void mme_s11_handle_delete_bearer_request(
      * If GTP-xact Holding timer is expired,
      * OLD bearer->xact memory will be automatically removed.
      */
-    bearer->delete.xact = xact;
+    ogs_assert(xact->id >= OGS_MIN_POOL_ID && xact->id <= OGS_MAX_POOL_ID);
+    bearer->delete.xact_id = xact->id;
 
     if (ECM_IDLE(mme_ue)) {
         MME_STORE_PAGING_INFO(mme_ue,
-            MME_PAGING_TYPE_DELETE_BEARER, bearer);
+            MME_PAGING_TYPE_DELETE_BEARER, bearer->id);
         r = s1ap_send_paging(mme_ue, S1AP_CNDomain_ps);
         ogs_expect(r == OGS_OK);
         ogs_assert(r != OGS_ERROR);
@@ -1234,7 +1374,6 @@ void mme_s11_handle_release_access_bearers_response(
     ogs_assert(rsp);
 
     ogs_debug("Release Access Bearers Response");
-    MME_UE_LIST_CHECK;
 
     /********************
      * Check Transaction
@@ -1242,7 +1381,8 @@ void mme_s11_handle_release_access_bearers_response(
     ogs_assert(xact);
     action = xact->release_action;
     ogs_assert(action);
-    mme_ue = mme_ue_cycle(xact->data);
+
+    mme_ue = mme_ue_find_by_id(OGS_POINTER_TO_UINT(xact->data));
 
     rv = ogs_gtp_xact_commit(xact);
     if (rv != OGS_OK) {
@@ -1254,7 +1394,7 @@ void mme_s11_handle_release_access_bearers_response(
         ogs_error("MME-UE Context has already been removed");
         return;
     }
-    sgw_ue = sgw_ue_cycle(mme_ue->sgw_ue);
+    sgw_ue = sgw_ue_find_by_id(mme_ue->sgw_ue_id);
     ogs_assert(sgw_ue);
 
     /***********************
@@ -1292,10 +1432,10 @@ void mme_s11_handle_release_access_bearers_response(
     }
 
     if (action == OGS_GTP_RELEASE_SEND_UE_CONTEXT_RELEASE_COMMAND) {
-        enb_ue = enb_ue_cycle(mme_ue->enb_ue);
+        enb_ue = enb_ue_find_by_id(mme_ue->enb_ue_id);
         if (enb_ue) {
             r = s1ap_send_ue_context_release_command(enb_ue,
-                    S1AP_Cause_PR_nas, S1AP_CauseNas_normal_release,
+                    enb_ue->relcause.group, enb_ue->relcause.cause,
                     S1AP_UE_CTX_REL_S1_REMOVE_AND_UNLINK, 0);
             ogs_expect(r == OGS_OK);
             ogs_assert(r != OGS_ERROR);
@@ -1320,17 +1460,18 @@ void mme_s11_handle_release_access_bearers_response(
      * for new UE-associated logical S1-connections over the S1 interface,
      * the MME shall respond with the RESET ACKNOWLEDGE message.
      */
-        enb_ue = enb_ue_cycle(mme_ue->enb_ue);
+        enb_ue = enb_ue_find_by_id(mme_ue->enb_ue_id);
 
         enb_ue_unlink(mme_ue);
 
         if (enb_ue) {
-            mme_enb_t *enb = enb_ue->enb;
-            ogs_assert(enb);
+            mme_enb_t *enb = NULL;
+
+            enb = mme_enb_find_by_id(enb_ue->enb_id);
 
             enb_ue_remove(enb_ue);
 
-            if (ogs_list_count(&enb->enb_ue_list) == 0) {
+            if (enb && ogs_list_count(&enb->enb_ue_list) == 0) {
                 r = s1ap_send_s1_reset_ack(enb, NULL);
                 ogs_expect(r == OGS_OK);
                 ogs_assert(r != OGS_ERROR);
@@ -1342,35 +1483,39 @@ void mme_s11_handle_release_access_bearers_response(
     } else if (action == OGS_GTP_RELEASE_S1_CONTEXT_REMOVE_BY_RESET_PARTIAL) {
         enb_ue_t *iter = NULL;
 
-        enb_ue = enb_ue_cycle(mme_ue->enb_ue);
+        enb_ue = enb_ue_find_by_id(mme_ue->enb_ue_id);
 
         enb_ue_unlink(mme_ue);
 
         if (enb_ue) {
-            mme_enb_t *enb = enb_ue->enb;
-            ogs_assert(enb);
+            mme_enb_t *enb = NULL;
+
+            enb = mme_enb_find_by_id(enb_ue->enb_id);
 
             enb_ue_remove(enb_ue);
 
-            ogs_list_for_each(&enb->enb_ue_list, iter) {
-                if (iter->part_of_s1_reset_requested == true) {
-                    /* The ENB_UE context
-                     * where PartOfS1_interface was requested
-                     * still remains */
-                    return;
+            if (enb) {
+                ogs_list_for_each(&enb->enb_ue_list, iter) {
+                    if (iter->part_of_s1_reset_requested == true) {
+                        /* The ENB_UE context
+                         * where PartOfS1_interface was requested
+                         * still remains */
+                        return;
+                    }
                 }
+
+                /* All ENB_UE context
+                 * where PartOfS1_interface was requested
+                 * REMOVED */
+                ogs_assert(enb->s1_reset_ack);
+                r = s1ap_send_to_enb(
+                        enb, enb->s1_reset_ack, S1AP_NON_UE_SIGNALLING);
+                ogs_expect(r == OGS_OK);
+                ogs_assert(r != OGS_ERROR);
+
+                /* Clear S1-Reset Ack Buffer */
+                enb->s1_reset_ack = NULL;
             }
-
-            /* All ENB_UE context
-             * where PartOfS1_interface was requested
-             * REMOVED */
-            ogs_assert(enb->s1_reset_ack);
-            r = s1ap_send_to_enb(enb, enb->s1_reset_ack, S1AP_NON_UE_SIGNALLING);
-            ogs_expect(r == OGS_OK);
-            ogs_assert(r != OGS_ERROR);
-
-            /* Clear S1-Reset Ack Buffer */
-            enb->s1_reset_ack = NULL;
         } else {
             ogs_error("ENB-S1 Context has already been removed");
         }
@@ -1405,7 +1550,7 @@ void mme_s11_handle_downlink_data_notification(
         ogs_error("No UE Context");
         cause_value = OGS_GTP2_CAUSE_CONTEXT_NOT_FOUND;
     } else {
-        sgw_ue = sgw_ue_cycle(mme_ue->sgw_ue);
+        sgw_ue = sgw_ue_find_by_id(mme_ue->sgw_ue_id);
         ogs_assert(sgw_ue);
 
         if (noti->eps_bearer_id.presence == 0) {
@@ -1448,7 +1593,8 @@ void mme_s11_handle_downlink_data_notification(
      * If GTP-xact Holding timer is expired,
      * OLD bearer->xact memory will be automatically removed.
      */
-    bearer->notify.xact = xact;
+    ogs_assert(xact->id >= OGS_MIN_POOL_ID && xact->id <= OGS_MAX_POOL_ID);
+    bearer->notify.xact_id = xact->id;
 
     if (noti->cause.presence) {
         ogs_gtp2_cause_t *cause = noti->cause.data;
@@ -1472,12 +1618,16 @@ void mme_s11_handle_downlink_data_notification(
  */
     if (ECM_IDLE(mme_ue)) {
         MME_STORE_PAGING_INFO(mme_ue,
-            MME_PAGING_TYPE_DOWNLINK_DATA_NOTIFICATION, bearer);
+            MME_PAGING_TYPE_DOWNLINK_DATA_NOTIFICATION, bearer->id);
         r = s1ap_send_paging(mme_ue, S1AP_CNDomain_ps);
         ogs_expect(r == OGS_OK);
         ogs_assert(r != OGS_ERROR);
     } else if (ECM_CONNECTED(mme_ue)) {
         MME_CLEAR_PAGING_INFO(mme_ue);
+        ogs_assert(OGS_OK ==
+            mme_gtp_send_downlink_data_notification_ack(
+                bearer, OGS_GTP2_CAUSE_UE_ALREADY_RE_ATTACHED));
+
         if (cause_value == OGS_GTP2_CAUSE_ERROR_INDICATION_RECEIVED) {
 
 /*
@@ -1508,7 +1658,7 @@ void mme_s11_handle_downlink_data_notification(
  *   included in Downlink Data Notification is "Error Indication received
  *   from RNC/eNodeB/S4-SGSN"
  */
-            enb_ue_t *enb_ue = enb_ue_cycle(mme_ue->enb_ue);
+            enb_ue_t *enb_ue = enb_ue_find_by_id(mme_ue->enb_ue_id);
             ogs_assert(enb_ue);
 
             r = s1ap_send_ue_context_release_command(enb_ue,
@@ -1516,10 +1666,6 @@ void mme_s11_handle_downlink_data_notification(
                     S1AP_UE_CTX_REL_S1_PAGING, 0);
             ogs_expect(r == OGS_OK);
             ogs_assert(r != OGS_ERROR);
-        } else {
-            ogs_assert(OGS_OK ==
-                mme_gtp_send_downlink_data_notification_ack(
-                    bearer, OGS_GTP2_CAUSE_UE_ALREADY_RE_ATTACHED));
         }
     }
 }
@@ -1546,7 +1692,7 @@ void mme_s11_handle_create_indirect_data_forwarding_tunnel_response(
      * Check Transaction
      ********************/
     ogs_assert(xact);
-    mme_ue = mme_ue_cycle(xact->data);
+    mme_ue = mme_ue_find_by_id(OGS_POINTER_TO_UINT(xact->data));
 
     rv = ogs_gtp_xact_commit(xact);
     if (rv != OGS_OK) {
@@ -1558,7 +1704,7 @@ void mme_s11_handle_create_indirect_data_forwarding_tunnel_response(
         ogs_error("MME-UE Context has already been removed");
         return;
     }
-    sgw_ue = sgw_ue_cycle(mme_ue->sgw_ue);
+    sgw_ue = sgw_ue_find_by_id(mme_ue->sgw_ue_id);
     ogs_assert(sgw_ue);
 
     /************************
@@ -1652,7 +1798,7 @@ void mme_s11_handle_create_indirect_data_forwarding_tunnel_response(
         }
     }
 
-    source_ue = enb_ue_cycle(mme_ue->enb_ue);
+    source_ue = enb_ue_find_by_id(mme_ue->enb_ue_id);
     ogs_assert(source_ue);
 
     r = s1ap_send_handover_command(source_ue);
@@ -1681,7 +1827,7 @@ void mme_s11_handle_delete_indirect_data_forwarding_tunnel_response(
     ogs_assert(xact);
     action = xact->delete_indirect_action;
     ogs_assert(action);
-    mme_ue = mme_ue_cycle(xact->data);
+    mme_ue = mme_ue_find_by_id(OGS_POINTER_TO_UINT(xact->data));
 
     rv = ogs_gtp_xact_commit(xact);
     if (rv != OGS_OK) {
@@ -1693,7 +1839,7 @@ void mme_s11_handle_delete_indirect_data_forwarding_tunnel_response(
         ogs_error("MME-UE Context has already been removed");
         return;
     }
-    sgw_ue = sgw_ue_cycle(mme_ue->sgw_ue);
+    sgw_ue = sgw_ue_find_by_id(mme_ue->sgw_ue_id);
     ogs_assert(sgw_ue);
 
     /************************
@@ -1761,7 +1907,7 @@ void mme_s11_handle_delete_indirect_data_forwarding_tunnel_response(
     if (action == OGS_GTP_DELETE_INDIRECT_HANDOVER_COMPLETE) {
         /* Nothing to do */
     } else if (action == OGS_GTP_DELETE_INDIRECT_HANDOVER_CANCEL) {
-        r = s1ap_send_handover_cancel_ack(mme_ue->enb_ue);
+        r = s1ap_send_handover_cancel_ack(enb_ue_find_by_id(mme_ue->enb_ue_id));
         ogs_expect(r == OGS_OK);
         ogs_assert(r != OGS_ERROR);
     } else {
@@ -1788,11 +1934,11 @@ void mme_s11_handle_bearer_resource_failure_indication(
      * Check Transaction
      ********************/
     ogs_assert(xact);
-    bearer = xact->data;
+    bearer = mme_bearer_find_by_id(OGS_POINTER_TO_UINT(xact->data));
     ogs_assert(ind);
-    sess = bearer->sess;
+    sess = mme_sess_find_by_id(bearer->sess_id);
     ogs_assert(sess);
-    mme_ue = mme_ue_cycle(sess->mme_ue);
+    mme_ue = mme_ue_find_by_id(sess->mme_ue_id);
 
     rv = ogs_gtp_xact_commit(xact);
     if (rv != OGS_OK) {
@@ -1804,7 +1950,7 @@ void mme_s11_handle_bearer_resource_failure_indication(
         ogs_error("MME-UE Context has already been removed");
         return;
     }
-    sgw_ue = sgw_ue_cycle(mme_ue->sgw_ue);
+    sgw_ue = sgw_ue_find_by_id(mme_ue->sgw_ue_id);
     ogs_assert(sgw_ue);
 
     if (!mme_ue_from_teid)
